package edu.udo.cs.uschi;

import edu.udo.cs.uschi.MyBeta;
import edu.udo.cs.uschi.MembershipVector;
import edu.udo.cs.uschi.scaler.TrivialScaler;
import edu.udo.cs.uschi.scaler.MinmaxScaler;
import edu.udo.cs.uschi.scaler.SoftmaxScaler;
import java.io.*;
import java.lang.Integer;
import java.util.Properties;

public class ScalingMethod
{
    int number_classes;
    char delimiter;
    MembershipVector memberships; // set of membership vectors

    TrivialScaler trivial_scaling;
    BetaScaler[] scaling; // the actual scaling functions

    Properties parameters;

    public ScalingMethod()
    {
	number_classes = 0;
	delimiter = ' ';
	memberships = new MembershipVector();
	scaling = null;
	parameters = new Properties();
	trivial_scaling = new TrivialScaler(parameters);
    };


    public ScalingMethod(Properties params)
	throws Exception
    {
	number_classes = 0;
	delimiter = ' ';
	memberships = new MembershipVector();
	scaling = null;
	parameters = params;
	String scaling_type = parameters.getProperty("scaling");
	if(scaling_type.equals("none")){
	    trivial_scaling = new TrivialScaler(parameters);
	}
	else if(scaling_type.equals("minmax")){
	    trivial_scaling = new MinmaxScaler(parameters);
	}
	else if(scaling_type.equals("softmax")){
	    trivial_scaling = new SoftmaxScaler(parameters);
	}
	else{
	    throw(new Exception("Unknown scaling type: "+scaling_type));
	};
    };


    protected void read_memberships(String filename)
	throws Exception
    {
	BufferedReader infile = new BufferedReader(new FileReader(filename));
	String line=null;
	boolean binary=false;
	double binary_threshold = Double.parseDouble(parameters.getProperty("binary_threshold"));
	// read first line and get the number of classes
	boolean ok;
	ok = false;
	while(!ok){
	    line = infile.readLine();
	    ok = true;
	    if(line != null){
		line = line.trim();
		if(line.length() == 0){
		    ok = false;
		}
		else if(line.charAt(0) == '#'){
		    ok = false;
		}
		else{
		    line.replace('\t',' ');
		};
	    };
	};
	if(line == null){
	    throw(new Exception("No examples found in file "+filename));
	};
	line.replace('\t',' ');
	number_classes= 0;
	for(int i=0; i<line.length(); i++){
	    if(line.charAt(i) == delimiter){
		number_classes++;
		    while(line.charAt(i) == ' '){
			// multiple spaces count as one delimiter
			i++;
		    };
		};
	};
	int to_be_read;
	if(number_classes == 1){
	    // we take m[1] = 1-m[0]
	    binary = true;
	    number_classes = 2;
	    to_be_read = 1;
	}
	else{
	    to_be_read = number_classes;
	};

	// init data structures
	memberships.clear();

	// read all lines
	int pos;
	int true_class;
	double[] membership_values;
	while(line != null){
	    // split line and store values
	    membership_values = new double[number_classes];
	    for(int i=0; i<to_be_read; i++){
	      pos = line.indexOf(delimiter);
	      if(pos < 0){ pos = line.length(); };
	      membership_values[i] = Double.parseDouble(line.substring(0,pos));
	      line = line.substring(pos+1).trim();
	    };
	    pos = line.indexOf(delimiter);
	    if(pos < 0){ pos = line.length(); };
	    true_class = Integer.parseInt(line.substring(0,pos));
	    true_class--; // Eingabe: 1..n, intern 0..n-1
	    if(binary){
		membership_values[1] = binary_threshold-membership_values[0];
		if(true_class != 0){
		    true_class = 1;
		};
	    };
	    memberships.add(true_class,membership_values);

	    // read next line
	    ok = false;
	    while(!ok){
		line = infile.readLine();
		ok = true;
		if(line != null){
		    line = line.trim();
		    if(line.length() == 0){
			ok = false;
		    }
		    else if(line.charAt(0) == '#'){
			ok = false;
		    }
		    else{
			line.replace('\t',' ');
		    };
		};
	    };
	};
	infile.close();
    };


    protected void write_scaled_memberships(String filename)
	throws Exception
    {
	BufferedWriter outfile = new BufferedWriter(new FileWriter(filename));
	int n=memberships.size();
	MembershipObject unscaled;
	MembershipObject scaled;
	double[] the_membership_values;
	int j;

	for(int i=0;i<n;i++){
	    unscaled = (MembershipObject)(memberships.get(i));
	    scaled = scale_membership(unscaled);
	    the_membership_values = scaled.memberships;
	    for(j=0;j<number_classes;j++){
		outfile.write(Double.toString(the_membership_values[j]));
		outfile.write(" ");
	    };
	    outfile.newLine();
	};
	outfile.close();	
    };


    protected MembershipObject scale_membership(MembershipObject unscaled)
	throws Exception
    {
	double[] scaled = new double[number_classes];
	int c = unscaled.predictedclass;
	scaled[c] = (scaling[c]).scale(unscaled.memberships[c]);
	if(scaled[c] == Double.NaN){
	    // numerical error, set scaled[c] to 1
	    scaled[c] = 1;
	};
	if(1-unscaled.memberships[c] == 0){
	    // scaled[i] = NaN
	    for(int i=0;i<number_classes;i++){
		if(i != c){
		    scaled[i] = (1-scaled[c])/(number_classes-1);
		};
	    };
	}
	else{
	    for(int i=0;i<number_classes;i++){
		if(i != c){
		    scaled[i] = (1-scaled[c])/(1-unscaled.memberships[c])*unscaled.memberships[i];
		};
	    };
	};
	MembershipObject scaled_memberships = new MembershipObject(unscaled.trueclass,scaled);
	return scaled_memberships;
    };


    protected void read_model(String filename)
	throws Exception
    {
	BufferedReader infile = new BufferedReader(new FileReader(filename));
	String line=null;

	// read first line
	boolean ok;
	ok = false;
	while(!ok){
	    line = infile.readLine();
	    ok = true;
	    if(line != null){
		line = line.trim();
		if(line.length() == 0){
		    ok = false;
		}
		else if(line.charAt(0) == '#'){
		    ok = false;
		};
	    };
	};
	if(line == null){
	    throw(new Exception("No model found in file "+filename));
	};
	number_classes= Integer.parseInt(line);

	// init data structures!!!
	scaling = new BetaScaler[number_classes];

	// read all lines
	int pos;
	int number_params = 4; // there are 4 parameters for each scaler
	double[] params = new double[number_params];

	for(int c=0;c<number_classes;c++){
	    // read next line
	    ok = false;
	    while(!ok){
		line = infile.readLine();
		ok = true;
		if(line != null){
		    line = line.trim();
		    if(line.length() == 0){
			ok = false;
		    }
		    else if(line.charAt(0) == '#'){
			ok = false;
		    };
		}
		else{
		    throw(new Exception("Not enough parameters in file "+filename));
		};
	    };
	    // split line and store values
	    for(int i=1; i<number_params; i++){
	      pos = line.indexOf(' ');
	      if(pos < 0){ pos = line.length(); };
	      params[i-1] = Double.parseDouble(line.substring(0,pos));
	      line = line.substring(pos+1).trim();
	    };
	    params[number_params-1] = Double.parseDouble(line);
	    scaling[c] = new BetaScaler(params);
	};
	trivial_scaling.read(infile);
	infile.close();
    };


    protected void write_model(String filename)
	throws Exception
    {
	BufferedWriter outfile = new BufferedWriter(new FileWriter(filename));
	int i;
	for(i=0;i<number_classes;i++){
	    outfile.write("# ");
	    outfile.write(Double.toString((scaling[i]).scale(0.5)));
	    outfile.newLine(); 
	};
	outfile.write("# number of classes");
	outfile.newLine(); 
	outfile.write(Integer.toString(number_classes));
	outfile.newLine(); 
	outfile.write("# beta parameters");
	outfile.newLine(); 	
	double[] params;
	for(int c=0;c<number_classes;c++){
	    outfile.write((scaling[c]).toString());
	    outfile.newLine();
	};
	trivial_scaling.write(outfile);
	outfile.close();
    };


    protected void generate_scaling()
	throws Exception
    {
	scaling = new BetaScaler[number_classes];
	long[] N_T = new long[number_classes];
	double[] avg_m = new double[number_classes];
	double[] var_m = new double[number_classes];
	double[] CR = new double[number_classes];
	MembershipObject the_membership;
	for(int c=0;c<number_classes;c++){
	    N_T[c]=0;
	    avg_m[c]=0;
	    var_m[c]=0;
	    CR[c] = 0;
	};
	int i=memberships.size();
	int c;
	double[] x;
	while(i>0){
	    i--;
	    the_membership = (MembershipObject)(memberships.elementAt(i));
	    c = the_membership.predictedclass;
	    x = the_membership.memberships;
	    N_T[c]++;
	    avg_m[c] += x[c];
	    var_m[c] += x[c]*x[c];
	    if(c == the_membership.trueclass){
		CR[c]++;
	    };
	};
	for(c=0;c<number_classes;c++){
	    if(N_T[c] < 2){
		N_T[c] = 2; // @@@ hack
	    };
	    // hacks to get consistent avg_m[c]
	    if(avg_m[c] == N_T[c]){
		avg_m[c] -= 1;
	    }
	    else if(avg_m[c] == 0){
		avg_m[c] += 1;
	    };
	    var_m[c]=(var_m[c]-avg_m[c]*avg_m[c]/N_T[c])/(N_T[c]-1);
	    avg_m[c]/=(double)(N_T[c]);
	    if(var_m[c]<= Double.parseDouble(parameters.getProperty("precision"))){ 
		// numerical error or constant var
		var_m[c] = Double.parseDouble(parameters.getProperty("precision"));
	    };
	    // hacks to get consistent CR[c]
	    if(CR[c] == N_T[c]){
		CR[c] -= 1;
	    }
	    else if(CR[c] == 0){
		CR[c] += 1;
	    };
	    CR[c] /= (double)N_T[c];


	};

	double p_m_c;
	double N_m_c;
	BetaScaler b_scaler;
	long Nmin;
	long Nmax;
	double D_N;
	long correctpredictions;
	double sqrdist;
	int trueclass;
	int predictedclass;
	int j;
	double min_D_N;
	long N_at_min_D_N;
	double scaled_m;
	double scaled_optimal_m;
	double scale_factor;
	long max_Nmax = Long.parseLong(parameters.getProperty("n_max"));
	for(c=0;c<number_classes;c++){
	    b_scaler = new BetaScaler(); // each class gets own scaler
	    p_m_c = avg_m[c];
	    N_m_c = (p_m_c*(1-p_m_c))/var_m[c]-1;
	    if(N_m_c < 1){
		N_m_c = 1;
	    };
	    if((double)N_T[c] < N_m_c){
		Nmin = N_T[c];
		Nmax = Math.round(N_m_c+1);
	    }
	    else{
		Nmin = Math.round(N_m_c-1);
		Nmax = N_T[c];
	    };
	    if(Nmin < 1){
		Nmin = 1;
	    };
	    if(Nmax > max_Nmax){
		Nmax = max_Nmax;
	    };
	    min_D_N = Double.POSITIVE_INFINITY;
	    N_at_min_D_N = 1;

	    for(long N=Nmin;N<=Nmax;N++){
		b_scaler.setMembershipParams(p_m_c,N_m_c);
		b_scaler.setTestsetParams(CR[c],N);
		D_N = 0;
		correctpredictions = 1;
		i = memberships.size();		
		while(i>0){
		    i--;
		    the_membership = (MembershipObject)(memberships.elementAt(i));
		    x = the_membership.memberships;
		    trueclass = the_membership.trueclass;
		    predictedclass = the_membership.predictedclass;
		    if(predictedclass == c){
			scaled_optimal_m = b_scaler.scale(x[predictedclass]);
			scale_factor = (1-scaled_optimal_m)/(1-x[predictedclass]);
			// calc \sum m^2(x)
			sqrdist = scaled_optimal_m*scaled_optimal_m;
			for(j=0;j<x.length;j++){
			    scaled_m = scale_factor*x[j];
			    sqrdist += scaled_m*scaled_m;
			};
			// summand of predictedclass is already in sum:
			scaled_m = scale_factor*x[predictedclass];
			sqrdist -= scaled_m*scaled_m;
			// 2nd step:
			// replace m^2[trueclass] with (1-m[trueclass])^2
			// i.e. + 1-2*m[trueclass]
			if(trueclass == predictedclass){
			    sqrdist += 1-2*scaled_optimal_m;
			}
			else{
			    scaled_m = scale_factor*x[trueclass];
			    sqrdist += 1-2*scaled_m;
			};
			if(sqrdist < 0){
			    // numerical error
			    sqrdist = 0;
			};
			if(((scaled_optimal_m > 0.5) && (predictedclass == trueclass)) ||
			   ((scaled_optimal_m <= 0.5) && (predictedclass != trueclass))){
			    correctpredictions++;
			}
			D_N += Math.sqrt(sqrdist);
		    };
		};
		D_N /= correctpredictions;
		if(D_N < min_D_N){
		    min_D_N = D_N;
		    N_at_min_D_N = N;
		};
	    };
	    b_scaler.setTestsetParams(CR[c],N_at_min_D_N);
	    scaling[c] = b_scaler;
	};
    };


    public void learn(String infile, String modelfile)
	throws Exception
    {
	read_memberships(infile);
	trivial_scaling.init(memberships);
	trivial_scaling.scale(memberships);
	generate_scaling();
	write_model(modelfile);
    };



    public void scale(String infile, String modelfile, String outfile)
	throws Exception
    {
	read_model(modelfile);            
	read_memberships(infile);         
	trivial_scaling.scale(memberships);
	write_scaled_memberships(outfile);
    };
};

